@@ -12,7 +12,7 @@ require 'shellwords'
 
 raise "sorry, 1.8 was last decade" unless RUBY_VERSION >= '1.9'
 
-gem 'kramdown', '~> 1.17.0'
+gem 'kramdown', '~> 2.3.0'
 require 'kramdown'
 my_span_elements =  %w{list figure xref eref iref cref spanx vspace}
 Kramdown::Parser::Html::Constants::HTML_SPAN_ELEMENTS.concat my_span_elements
@@ -34,38 +34,82 @@ module Kramdown
 
     class RFC2629Kramdown < Kramdown
 
+      def replace_abbreviations(el, regexps = nil)
+        unless regexps          # DUPLICATED AND MODIFIED CODE FROM UPSTREAM, CHECK ON UPSTREAM UPGRADE
+          sorted_abbrevs = @root.options[:abbrev_defs].keys.sort {|a, b| b.length <=> a.length }
+          regexps = [Regexp.union(*sorted_abbrevs.map {|k| 
+                                    /#{Regexp.escape(k).gsub(/\\\s/, "[\\s\\p{Z}]+").force_encoding(Encoding::UTF_8)}/})]
+          # warn regexps.inspect
+          regexps << /(?=(?:\W|^)#{regexps.first}(?!\w))/ # regexp should only match on word boundaries
+        end
+        super(el, regexps)
+      end
+
+
       def initialize(*doc)
         super
         @span_parsers.unshift(:xref)
         @span_parsers.unshift(:iref)
+        @span_parsers.unshift(:span_pi)
+        @block_parsers.unshift(:block_pi)
       end
 
-      SECTIONS_RE = /Section(?:s (?:[\w.]+, )*[\w.]+,? and)? [\w.]+/
+      XREF_BASE = /#{REXML::XMLTokens::NAME_CHAR}+/ # a token for a reference
+      XREF_TXT = /(?:[^\(]|\([^\)]*\))+/ # parenthesized text
+      XREF_RE = /#{XREF_BASE}(?: \(#{XREF_TXT}\))?/
+      XREF_RE_M = /\A(#{XREF_BASE})(?: \((#{XREF_TXT})\))?/ # matching version of XREF_RE
+      XREF_SINGLE = /(?:Section|Appendix) #{XREF_RE}/
+      XREF_MULTI = /(?:Sections|Appendices) (?:#{XREF_RE}, )*#{XREF_RE},? and #{XREF_RE}/
+      XREF_ANY = /(?:#{XREF_SINGLE}|#{XREF_MULTI})/
+      SECTIONS_RE = /(?:#{XREF_ANY} and )?#{XREF_ANY}/
 
-      def handle_bares(s, attr, format, href)
-        sa = s.sub(/\A\S+\s/, '').split(/,? and /)
-        sa[0..0] = *sa[0].split(', ')
-        sz = sa.size
-        if sz != 1         # we have to redo xml2rfc's work here
-          @tree.children << Element.new(:text, "Sections ", {}) # XXX needs to split into Section/Appendix
-          sa.each_with_index do |sec, i|
-            attr1 = {"target" => href, "section" => sec, "sectionFormat" => "bare"}
-            @tree.children << Element.new(:xref, nil, attr1)
-            text = if i == 0 && sz == 2
-                     " and "
-                   elsif i == sz-1
-                     " of "
-                   elsif i == sz-2
-                     ", and "
-                   else
-                     ", "
-                   end
-            @tree.children << Element.new(:text, text, {})
+      def self.idref_cleanup(href)
+        # can't start an IDREF with a number or reserved start
+        if href =~ / /
+          if $options.v3
+            warn "** space(s) in cross-reference '#{href}' -- are you trying to use section references?"
+          else
+            warn "** space(s) in cross-reference '#{href}' -- note that section references are supported in v3 mode only."
           end
-          # attr stays unchanged, no section added
-        else
-          attr['section'] = sa[-1]
-          attr['sectionFormat'] = format
+        end
+        href.gsub(/\A(?:[0-9]|section-|u-|figure-|table-|iref-)/) { "_#{$&}" }
+      end
+
+      def handle_bares(s, attr, format, href, last_join = nil)
+        if s.match(/\A(#{XREF_ANY}) and (#{XREF_ANY})\z/)
+          handle_bares($1, {}, nil, href, " and ")
+          handle_bares($2, {}, nil, href, " of ")
+          return
+        end
+
+        href = href.split(' ')[0] # Remove any trailing (...)
+        multi = last_join != nil
+        (sn, s) = s.split(' ', 2)
+        loop do
+          m = s.match(/\A#{XREF_RE_M}(, (?:and )?| and )?/)
+          break if not m
+
+          if not multi and not m[2] and not m[3]
+            # Modify |attr| if there is a single reference.  This can only be
+            # used if there is only one section reference and the section part
+            # has no title.
+            attr['section'] = m[1]
+            attr['sectionFormat'] = format
+            attr['text'] = m[2]
+            return
+          end
+
+          if sn
+            @tree.children << Element.new(:text, "#{sn} ", {})
+            sn = nil
+          end
+
+          multi = true
+          s[m[0]] = ''
+
+          attr1 = { 'target' => href, 'section' => m[1], 'sectionFormat' => 'bare', 'text' => m[2] }
+          @tree.children << Element.new(:xref, nil, attr1)
+          @tree.children << Element.new(:text, m[3] || last_join || " of ", {})
         end
       end
 
@@ -100,19 +144,34 @@ module Kramdown
             when /\A(.*) \((#{SECTIONS_RE})\)\z/
               href = $1
               handle_bares($2, attr, "parens", href)
-            when /\A([\w.]+)<(.*)\z/
-              href = $2
-              attr['section'] = $1
+            when /#{XREF_RE_M}<(.+)\z/
+              href = $3
+              if $2
+                attr['section'] = $2
+                attr['relative'] = "#" << $1
+              else
+                attr['section'] = $1
+              end
               attr['sectionFormat'] = 'bare'
+            when /\A<<(.+)\z/
+              href = $1
+              attr['format'] = 'title'
+            when /\A<(.+)\z/
+              href = $1
+              attr['format'] = 'counter'
             end
           end
-          href = href.gsub(/\A[0-9]/) { "_#{$&}" } # can't start an IDREF with a number
+          if href.match(/#{XREF_RE_M}\z/)
+            href = $1
+            attr['text'] = $2
+          end
+          href = self.class.idref_cleanup(href)
           attr['target'] = href
           el = Element.new(:xref, nil, attr)
         end
         @tree.children << el
       end
-      define_parser(:xref, XREF_START, '{{')
+      define_parser(:xref, XREF_START, '\{\{')
 
       IREF_START = /\(\(\((.*?)\)\)\)/u
 
@@ -125,6 +184,40 @@ module Kramdown
       end
       define_parser(:iref, IREF_START, '\(\(\(')
 
+      # HTML_INSTRUCTION_RE = /<\?(.*?)\?>/m    # still defined!
+
+      # warn [:OPT_SPACE, OPT_SPACE, HTML_INSTRUCTION_RE].inspect
+
+      PI_BLOCK_START = /^#{OPT_SPACE}<\?/u
+
+      def parse_block_pi
+        # warn [:BLOCK].inspect
+        line = @src.current_line_number
+        if (result = @src.scan(HTML_INSTRUCTION_RE))
+          @tree.children << Element.new(:xml_pi, result, nil, category: :block, location: line)
+          @src.scan(TRAILING_WHITESPACE)
+          true
+        else
+          false
+        end
+      end
+      define_parser(:block_pi, PI_BLOCK_START)
+
+      PI_SPAN_START = /<\?/u
+
+      def parse_span_pi
+        # warn [:SPAN].inspect
+        line = @src.current_line_number
+        if (result = @src.scan(HTML_INSTRUCTION_RE))
+          @tree.children << Element.new(:xml_pi, result, nil, category: :span, location: line)
+        else
+          add_text(@src.getch)
+        end
+      end
+      define_parser(:span_pi, PI_SPAN_START, '<\?')
+
+      # warn [:HERE, @@parsers.keys].inspect
+
     end
   end
 
@@ -145,13 +238,20 @@ module Kramdown
     )
     STUDLY_ATTR_MAP = Hash[STUDLY_ATTR.map {|s| [s.downcase, s]}]
 
-    def rfc2629_fix
+    TRUTHY = Hash.new {|h, k| k}
+    TRUTHY["false"] = false
+    TRUTHY["no"] = false
+
+    def rfc2629_fix(opts)
       if a = attr
         if anchor = a.delete('id')
-          a['anchor'] = anchor
+          a['anchor'] = ::Kramdown::Parser::RFC2629Kramdown.idref_cleanup(anchor)
         end
         if anchor = a.delete('href')
-          a['target'] = anchor
+          a['target'] = ::Kramdown::Parser::RFC2629Kramdown.idref_cleanup(anchor)
+        end
+        if av = a.delete('noabbrev')      # pseudo attribute -> opts
+          opts = opts.merge(noabbrev: TRUTHY[av]) # updated copy
         end
         attr.keys.each do |k|
           if (d = k.gsub(/_(.|$)/) { $1.upcase }) != k or d = STUDLY_ATTR_MAP[k]
@@ -159,6 +259,7 @@ module Kramdown
           end
         end
       end
+      opts
     end
   end
 
@@ -187,7 +288,10 @@ module Kramdown
           require 'net/http/persistent'
           $http = Net::HTTP::Persistent.new name: 'kramdown-rfc'
         rescue Exception => e
-          warn "** Can't set up persistent HTTP -- #{e}"
+          warn "** Not using persistent HTTP -- #{e}"
+          warn "**   To silence this message and get full speed, try:"
+          warn "**     gem install net-http-persistent"
+          warn "**   If this doesn't work, you can ignore this warning."
         end
       end
 
@@ -210,15 +314,15 @@ module Kramdown
       end
 
       def convert1(el, indent, opts = {})
-        el.rfc2629_fix
-        send("convert_#{el.type}", el, indent, opts)
+        nopts = el.rfc2629_fix(opts)
+        send("convert_#{el.type}", el, indent, nopts)
       end
 
       def inner_a(el, indent, opts)
         indent += INDENTATION
         el.children.map do |inner_el|
-          inner_el.rfc2629_fix
-          send("convert_#{inner_el.type}", inner_el, indent, opts)
+          nopts = inner_el.rfc2629_fix(opts)
+          send("convert_#{inner_el.type}", inner_el, indent, nopts)
         end
       end
 
@@ -246,6 +350,12 @@ module Kramdown
         generate_id(value).gsub(/-+/, '-')
       end
 
+      def self.process_markdown(v)             # Uuh.  Heavy coupling.
+        doc = ::Kramdown::Document.new(v, $global_markdown_options)
+        $stderr.puts doc.warnings.to_yaml unless doc.warnings.empty?
+        doc.to_rfc2629[3..-6] # skip <t>...</t>\n
+      end
+
       SVG_COLORS = Hash.new {|h, k| k}
       <<COLORS.each_line {|l| k, v = l.chomp.split; SVG_COLORS[k] = v}
 black	#000000
@@ -293,6 +403,31 @@ COLORS
         end
       end
 
+      SVG_NAMESPACES = {"xmlns"=>"http://www.w3.org/2000/svg",
+                        "xlink"=>"http://www.w3.org/1999/xlink"}
+
+      def svg_clean_kgt(s)
+        d = REXML::Document.new(s)
+        REXML::XPath.each(d.root, "/xmlns:svg", SVG_NAMESPACES) do |x|
+          if (w = x.attributes["width"]) && (h = x.attributes["height"])
+            x.attributes["viewBox"] = "0 0 %d %d" % [w, h]
+          end
+          if x.attributes["viewBox"]
+            x.attributes["width"] = nil
+            x.attributes["height"] = nil
+          end
+        end
+        REXML::XPath.each(d.root, "//rect|//line|//path") do |x|
+          x.attributes["fill"] = "none"
+          x.attributes["stroke"] = "black"
+          x.attributes["stroke-width"] = "1.5"
+        end
+        d.to_s
+      rescue => detail
+        warn "*** Can't clean SVG: #{detail}"
+        d
+      end
+
       def svg_clean(s)          # expensive, risky
         d = REXML::Document.new(s)
         REXML::XPath.each(d.root, "//*[@shape-rendering]") { |x| x.attributes["shape-rendering"] = nil }  #; warn x.inspect }
@@ -332,43 +467,84 @@ COLORS
         end
       end
 
-      def svg_tool_process(t, result)
+      def shell_prepare(opt)
+        " " << opt.shellsplit.shelljoin
+      end
+
+      DEFAULT_AASVG="aasvg --spaces=1"
+
+      def svg_tool_process(t, svg_opt, txt_opt, result)
         require 'tempfile'
         file = Tempfile.new("kramdown-rfc")
         file.write(result)
         file.close
+        dont_clean = false
+        dont_check = false
+        svg_opt = shell_prepare(svg_opt) if svg_opt
+        txt_opt = shell_prepare(txt_opt) if txt_opt
         case t
+        when "protocol", "protocol-goat", "protocol-aasvg"
+          cmdparm = result.lines.map(&:strip).select {|x| x != ''}.join(',')
+          result, err, _s = Open3.capture3("protocol #{Shellwords.escape(cmdparm)}#{txt_opt}", stdin_data: '')
+          if t == "protocol-goat"
+            file.unlink
+            file = Tempfile.new("kramdown-rfc")
+            file.write(result)
+            file.close
+            result1, err, _s = Open3.capture3("goat#{svg_opt} #{file.path}", stdin_data: result);
+            dont_clean = true
+          elsif t == "protocol-aasvg"
+            result1, err, _s = Open3.capture3("#{DEFAULT_AASVG}#{svg_opt}", stdin_data: result);
+            dont_clean = true
+          else
+            result1 = nil
+          end
         when "goat"
-          result1, err, _s = Open3.capture3("goat #{file.path}", stdin_data: result);
+          result1, err, _s = Open3.capture3("goat#{svg_opt} #{file.path}", stdin_data: result);
+          dont_clean = true
+        when "aasvg"
+          result1, err, _s = Open3.capture3("#{DEFAULT_AASVG}#{svg_opt}", stdin_data: result);
+          dont_clean = true
         when "ditaa"        # XXX: This needs some form of option-setting
-          result1, err, _s = Open3.capture3("ditaa #{file.path} --svg -o -", stdin_data: result);
+          result1, err, _s = Open3.capture3("ditaa #{file.path} --svg -o -#{svg_opt}", stdin_data: result);
         when "mscgen"
-          result1, err, _s = Open3.capture3("mscgen -T svg -i #{file.path} -o -", stdin_data: result);
+          result1, err, _s = Open3.capture3("mscgen -T svg -i #{file.path} -o -#{svg_opt}", stdin_data: result);
         when "mermaid"
-          result1, err, _s = Open3.capture3("mmdc -i #{file.path}", stdin_data: result); #  -b transparent
+          result1, err, _s = Open3.capture3("mmdc -i #{file.path}#{svg_opt}", stdin_data: result); #  -b transparent
           outpath = file.path + ".svg"
           result1 = File.read(outpath) rescue '' # don't die before providing error message
           File.unlink(outpath) rescue nil        # ditto
         when "plantuml", "plantuml-utxt"
           plantuml = "@startuml\n#{result}\n@enduml"
-          result1, err, _s = Open3.capture3("plantuml -pipe -tsvg", stdin_data: plantuml);
-          result, err1, _s = Open3.capture3("plantuml -pipe -tutxt", stdin_data: plantuml) if t == "plantuml-utxt"
+          result1, err, _s = Open3.capture3("plantuml -pipe -tsvg#{svg_opt}", stdin_data: plantuml);
+          result, err1, _s = Open3.capture3("plantuml -pipe -tutxt#{txt_opt}", stdin_data: plantuml) if t == "plantuml-utxt"
+          err << err1.to_s
+        when "railroad", "railroad-utf8"
+          result1, err1, _s = Open3.capture3("kgt -l abnf -e svg#{svg_opt}", stdin_data: result);
+          result1 = svg_clean_kgt(result1); dont_clean = true
+          result, err, _s = Open3.capture3("kgt -l abnf -e rr#{t == "railroad" ? "text" : "utf8"}#{txt_opt}",
+                                            stdin_data: result);
           err << err1.to_s
         when "math"
-          result1, err, _s = Open3.capture3("tex2svg --font STIX --speech=false #{Shellwords.escape(' ' << result)}");
-          result, err1, _s = Open3.capture3("asciitex -f #{file.path}")
+          result1, err, _s = Open3.capture3("tex2svg --font STIX --speech=false#{svg_opt} #{Shellwords.escape(' ' << result)}");
+          result, err1, _s = Open3.capture3("asciitex -f #{file.path}#{txt_opt}")
           err << err1
         end
         capture_croak(t, err)
-        # warn ["goat:", result1.inspect]
+        # warn ["text:", result.inspect]
+        # warn ["svg:", result1.inspect]
         file.unlink
-        result1 = svg_clean(result1) unless t == "goat"
-        result1, err, _s = Open3.capture3("svgcheck -qa", stdin_data: result1);
-        capture_croak("svgcheck", err)
-        # warn ["svgcheck:", result1.inspect]
-        if result1 == ''
-          warn "*** could not create svg for #{result.inspect[0...20]}..."
-          exit 65 # EX_DATAERR
+        if result1
+          result1 = svg_clean(result1) unless dont_clean
+          unless dont_check
+            result1, err, _s = Open3.capture3("svgcheck -Xqa", stdin_data: result1);
+            # warn ["svgcheck:", result1.inspect]
+            capture_croak("svgcheck", err)
+          end
+          if result1 == ''
+            warn "*** could not create svg for #{result.inspect[0...20]}..."
+            exit 65 # EX_DATAERR
+          end
         end
         [result, result1]       # text, svg
       end
@@ -420,12 +596,26 @@ COLORS
             end
           end
           case t
-          when "goat", "ditaa", "mscgen", "plantuml", "plantuml-utxt", "mermaid", "math"
+          when "aasvg", "ditaa", "goat",
+               "math", "mermaid",  "mscgen",
+               "plantuml", "plantuml-utxt",
+               "protocol", "protocol-aasvg", "protocol-goat",
+               "railroad", "railroad-utf8"
             if gi
               warn "*** Can't set GI #{gi} for composite SVG artset"
             end
-            result, result1 = memoize(:svg_tool_process, t, result)
-            "#{' '*indent}<figure#{el_html_attributes(el)}><artset><artwork #{html_attributes(artwork_attr.merge("type"=> "svg"))}>#{result1.sub(/.*?<svg/m, "<svg")}</artwork><artwork #{html_attributes(artwork_attr.merge("type"=> "ascii-art"))}><![CDATA[#{result}#{result =~ /\n\Z/ ? '' : "\n"}]]></artwork></artset></figure>\n"
+            result, result1 = memoize(:svg_tool_process, t,
+                                      artwork_attr.delete("svg-options"),
+                                      artwork_attr.delete("txt-options"),
+                                      result)
+            retart = mk_artwork(artwork_attr, "ascii-art",
+                                "<![CDATA[#{result}#{result =~ /\n\Z/ ? '' : "\n"}]]>")
+            if result1          # nest TXT in artset with SVG
+              retsvg = mk_artwork(artwork_attr, "svg",
+                                  result1.sub(/.*?<svg/m, "<svg"))
+              retart = "<artset>#{retsvg}#{retart}</artset>"
+            end
+            "#{' '*indent}<figure#{el_html_attributes(el)}>#{retart}</figure>\n"
           else
             gi ||= (
               if !$options.v3 || !t || ARTWORK_TYPES.include?(t) || artwork_attr["align"]
@@ -434,15 +624,57 @@ COLORS
                 "sourcecode"
               end
             )
+            loc_str =
+              if anchor = el.attr['anchor']
+                "##{anchor}"
+              elsif lineno = el.options[:location]
+                "approx. line #{lineno}" # XXX
+              else
+                "UNKNOWN"
+              end
+            case t
+            when "json"
+              begin
+                JSON.load(result)
+              rescue => e
+                err1 = "*** #{loc_str}: JSON isn't: #{e.message[0..40]}\n"
+                begin
+                  JSON.load("{" << result << "}")
+                rescue => e
+                  warn err1 << "***  not even with braces added around: #{e.message[0..40]}"
+                end
+              end
+            when "json-from-yaml"
+              begin
+                y = YAML.safe_load(result, aliases: true, filename: loc_str)
+                result = JSON.pretty_generate(y)
+                t = "json"      # XXX, this could be another format!
+              rescue => e
+                warn "*** YAML isn't: #{e.message}\n"
+              end
+            end
             "#{' '*indent}<figure#{el_html_attributes(el)}><#{gi}#{html_attributes(artwork_attr)}><![CDATA[#{result}#{result =~ /\n\Z/ ? '' : "\n"}]]></#{gi}></figure>\n"
           end
         end
       end
 
+      def mk_artwork(artwork_attr, typ, content)
+        "<artwork #{html_attributes(artwork_attr.merge("type" => typ))}>#{content}</artwork>"
+      end
+
       def convert_blockquote(el, indent, opts)
         text = inner(el, indent, opts)
+        if $options.v3
+          gi = el.attr.delete('gi')
+          if gi && gi != 'ul'
+            "#{' '*indent}<#{gi}#{el_html_attributes(el)}>\n#{text}#{' '*indent}</#{gi}>\n"
+          else
+            "#{' '*indent}<ul#{el_html_attributes_with(el, {"empty" => 'true'})}><li>\n#{text}#{' '*indent}</li></ul>\n"
+          end
+        else
         text = "<t></t>" unless text =~ /</ # empty block quote
         "#{' '*indent}<t><list style='empty'#{el_html_attributes(el)}>\n#{text}#{' '*indent}</list></t>\n"
+        end
       end
 
       def end_sections(to_level, indent)
@@ -481,7 +713,8 @@ COLORS
           if sl = el.attr.delete('slugifiedName') # could do general name- play
             attrstring = html_attributes({'slugifiedName' => sl})
           end
-          irefs = "<name#{attrstring}>#{inner(el, indent, opts)}</name>" #
+          # noabbrev: true -- Workaround for https://trac.ietf.org/trac/xml2rfc/ticket/683
+          irefs = "<name#{attrstring}>#{inner(el, indent, opts.merge(noabbrev: true))}</name>" #
         else
         clean, irefs = clean_pcdata(inner_a(el, indent, opts))
         el.attr['title'] = clean
@@ -662,7 +895,11 @@ COLORS
       end
 
       def convert_br(el, indent, opts)
-        "<vspace />"
+        if $options.v3
+          "<br />"
+        else
+          "<vspace />"
+        end
       end
 
       def convert_a(el, indent, opts)
@@ -683,6 +920,7 @@ COLORS
 
       def convert_xref(el, indent, opts)
         gi = el.attr.delete('gi')
+        text = el.attr.delete('text')
         target = el.attr['target']
         if target[0] == "&"
           "#{target};"
@@ -692,7 +930,12 @@ COLORS
           else
             gi ||= "xref"
           end
-          "<#{gi}#{el_html_attributes(el)}/>"
+          if text
+            tail = ">#{Rfc2629::process_markdown(text)}</#{gi}>"
+          else
+            tail = "/>"
+          end
+          "<#{gi}#{el_html_attributes(el)}#{tail}"
         end
       end
 
@@ -787,16 +1030,31 @@ COLORS
         end
       end
 
+      def self.bcp_std_ref(t, n)
+        warn "*** #{t} anchors not supported in v2 format" unless $options.v3
+        [name = "reference.#{t}.#{"%04d" % n.to_i}.xml",
+         "#{XML_RESOURCE_ORG_PREFIX}/bibxml-rfcsubseries-new/#{name}"] # FOR NOW
+      end
+
       # [subdirectory name, cache ttl in seconds, does it provide for ?anchor=]
       XML_RESOURCE_ORG_MAP = {
         "RFC" => ["bibxml", 86400*7, false,
-                  ->(fn, n){ "https://www.rfc-editor.org/refs/bibxml/#{fn}"}
+                  ->(fn, n){ [name = "reference.RFC.#{"%04d" % n.to_i}.xml",
+                              "https://www.rfc-editor.org/refs/bibxml/#{name}"] }
                  ],
         "I-D" => ["bibxml3", false, false,
-                  ->(fn, n){ "https://datatracker.ietf.org/doc/bibxml3/draft-#{n.sub(/\Adraft-/, '')}/xml" }
+                  ->(fn, n){ [fn,
+                              "https://datatracker.ietf.org/doc/bibxml3/draft-#{n.sub(/\Adraft-/, '')}.xml"] }
+                 ],
+        "BCP" => ["bibxml-rfcsubseries", 86400*7, false,
+                  ->(fn, n){ Rfc2629::bcp_std_ref("BCP", n) }
+                 ],
+        "STD" => ["bibxml-rfcsubseries", 86400*7, false,
+                  ->(fn, n){ Rfc2629::bcp_std_ref("STD", n) }
                  ],
         "W3C" => "bibxml4",
         "3GPP" => "bibxml5",
+        "SDO-3GPP" => "bibxml5",
         "ANSI" => "bibxml2",
         "CCITT" => "bibxml2",
         "FIPS" => "bibxml2",
@@ -835,16 +1093,16 @@ COLORS
             warn "*** missing anchor for '#{src}'"
             src
           )
-          anchor.sub!(/\A[0-9]/) { "_#{$&}" } # can't start an ID with a number
+          anchor = ::Kramdown::Parser::RFC2629Kramdown.idref_cleanup(anchor)
           anchor.gsub!('/', '_')              # should take out all illegals
           to_insert = ""
-          src.scan(/(W3C|3GPP|[A-Z-]+)[.]?([A-Za-z_0-9.\/\+-]+)/) do |t, n|
+          src.scan(/(W3C|3GPP|[A-Z-]+)[.]?([A-Za-z_0-9.\(\)\/\+-]+)/) do |t, n|
             fn = "reference.#{t}.#{n}.xml"
             sub, ttl, _can_anchor, altproc = XML_RESOURCE_ORG_MAP[t]
             ttl ||= KRAMDOWN_REFCACHETTL  # everything but RFCs might change a lot
             puts "*** Huh: #{fn}" unless sub
             if altproc && !KRAMDOWN_USE_TOOLS_SERVER
-              url = altproc.call(fn, n)
+              fn, url = altproc.call(fn, n)
             else
               url = "#{XML_RESOURCE_ORG_PREFIX}/#{sub}/#{fn}"
             end
@@ -867,10 +1125,12 @@ COLORS
                   }
                 else
                   REXML::XPath.each(d.root, "/reference/format") { |x|
-                    x.attributes["target"].sub!(%r{https?://www.ietf.org/internet-drafts/}, 
+                    x.attributes["target"].sub!(%r{https?://www.ietf.org/internet-drafts/},
                                                 %{https://www.ietf.org/archive/id/}) if t == "I-D"
                   }
                 end
+              elsif t == "IANA"
+                d.root.attributes["target"].sub!(%r{\Ahttp://www.iana.org/assignments/}, 'https://www.iana.org/assignments/')
               end
               to_insert = d.to_s
             rescue Exception => e
@@ -903,18 +1163,30 @@ COLORS
         # this would be more like xml2rfc v3:
         # "\n#{' '*indent}<cref>\n#{inner(el.value, indent, opts).rstrip}\n#{' '*indent}</cref>"
         content = inner(el.value, indent, opts).strip
-        content = escape_html(content.sub(/\A<t>(.*)<\/t>\z/m) {$1}, :text) # text only...
-        name = el.options[:name].sub(/\A[0-9]/) {"_" << $&}
+        content = content.sub(/\A<t>(.*)<\/t>\z/m) {$1}
+        name = ::Kramdown::Parser::RFC2629Kramdown.idref_cleanup(el.options[:name])
+        o_name = name.dup
         while @footnote_names_in_use[name] do
-          if name =~ /:\d+\z/
+          if name =~ /_\d+\z/
             name.succ!
           else
-            name << ":1"
+            name << "_1"
           end
         end
         @footnote_names_in_use[name] = true
         attrstring = el_html_attributes_with(el, {"anchor" => name})
-        "\n#{' '*indent}<cref#{attrstring}>#{content}</cref>"
+        if $options.v3
+          if o_name[-1] == "-"
+            # Ignore HTML attributes.  Hmm.
+            content
+          else
+            # do not indent span-level so we can stick to previous word.  Good?
+            "<cref#{attrstring}>#{content}</cref>"
+          end
+        else
+          content = escape_html(content, :text) # text only...
+          "\n#{' '*indent}<cref#{attrstring}>#{content}</cref>"
+        end
       end
 
       def convert_raw(el, indent, opts)
@@ -1029,9 +1301,34 @@ COLORS
       end
 
       def convert_abbreviation(el, indent, opts) # XXX: This is wrong
-        title = @root.options[:abbrev_defs][el.value]
-        title = nil if title.empty?
+        if opts[:noabbrev]
+          return el.value
+        end
+
         value = el.value
+        ix = value.gsub(/[\s\p{Z}]+/, " ")
+        title = @root.options[:abbrev_defs][ix]
+        if title.nil?
+          warn "*** abbrev mismatch: value = #{value.inspect} ix = #{ix.inspect}"
+        else
+          title = nil if title.empty?
+        end
+
+        if title == "<bcp14>" && $options.v3
+          return "<bcp14>#{value}</bcp14>"
+        end
+
+        if title && title[0] == "#"
+          target, title = title.split(' ', 2)
+          if target == "#"
+            target = value
+          else
+            target = target[1..-1]
+          end
+        else
+          target = nil
+        end
+
         if item = title
           m = title.scan(Parser::RFC2629Kramdown::IREF_START)
           if m.empty?
@@ -1043,7 +1340,11 @@ COLORS
           item = value
         end
         iref ||= "<iref#{html_attributes(item: item, subitem: subitem)}/>"
-        "#{el.value}#{iref}"
+        if target
+          "<xref#{html_attributes(target: target, format: "none")}>#{value}</xref>#{iref}"
+        else
+          "#{value}#{iref}"
+        end
       end
 
       def convert_root(el, indent, opts)
